local addonName, addonTable = ...
local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]

local name, _, _, enabled = GetAddOnInfo("tekDebug")
local player = UnitName("player")
local enabled = GetAddOnEnableState(player, "tekDebug") == 2
if enabled and not IsAddOnLoaded("tekDebug") then
  local succ, err = LoadAddOn(name)
end
local debugf = debug and tekDebug and tekDebug:GetFrame(addonName)
local function DebugPrint(...) if debugf then debugf:AddMessage(string.join(", ", tostringall(...))) end end

local strfind, strformat, strgsub, strlower, strmatch, tostring = string.find, string.format, string.gsub, string.lower, strmatch, tostring
local _, date, ipairs, next, pairs, select, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, time, table.insert, table.sort, type, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local PTLDB
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local PT = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale(addonName, false)
local DB_VERSION = 1
local currentRealmChar, key, id

if debug then
  ProfessionTrack = PT
end

local RAID_CLASS_COLORS_FONTS = {}


local function setColor(info, r, g, b, a)
  PT.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
end

local function getColor(info)
  return PT.db.profile[info[#info]].r, PT.db.profile[info[#info]].g, PT.db.profile[info[#info]].b, PT.db.profile[info[#info]].a
end


local defaults = {
  profile = {
    usedcd = { r=1, g=0, b=0, a=1 },
    freecd = { r=0, g=1, b=0, a=1 },
    tooltipBGColor            = { r=0, g=0, b=0, a=1 },
    useClassColors            = true,
    useCustomClassColors      = true,
    showLDBCooldownsAvailable = true,
    showLDBCooldownsText      = "CD: %d",
    showLDBItemsAvailable     = true,
    showRealmHeaderLines      = false,
    showRealmSumLines         = false,
    showCooldowns             = true,
    showBooks                 = false,
    minLevel = 1,
    minimapIcon = {
      hide = true,
      minimapPos = 220,
      radius = 80,
    },
  },
  global = {
    roster = {},
    cooldowns = {},
    items = {},
    trackedItems = {
      [108257] = true, -- Truesteel Ingot
      [108996] = true, -- Alchemical Catalyst
      [110611] = true, -- Burnished Leather
      [111366] = true, -- Gearspring Parts
      [111556] = true, -- Hexweave Cloth
      [112377] = true, -- War Paints
      [115524] = true, -- Taladite Crystal
      [115504] = true, -- Fractured Temporal Crystal
    },
    trackedSort = {
      [1]      = 108257,
      [108257] = 108996,
      [108996] = 110611,
      [110611] = 111366,
      [111366] = 111556,
      [111556] = 112377,
      [112377] = 115524,
      [115524] = 115504,
      [115504] = 0,
    },
  },
}

PT.professionCooldowns = {
  -- skillLine is the one returned by GetProfessionInfo( ... )
  -- need to check if any of multiple spells is on cooldown
  [171] = { iconid = 108996, spells = { 156587, 168042 }, book = 175880 }, -- alchemy
  [164] = { iconid = 108257, spells = { 171690, 171718 }, book = 176090 }, -- blacksmithing
  [333] = { iconid = 115504, spells = { 169092, 178241 }, book = 177043 }, -- enchating
  [202] = { iconid = 111366, spells = { 169080, 178242 }, book = 177054 }, -- engineering
  [773] = { iconid = 112377, spells = { 169081, 178240 }, book = 177045 }, -- inscription
  [755] = { iconid = 115524, spells = { 170700, 170832 }, book = 176087 }, -- jewelcrafting
  [165] = { iconid = 110611, spells = { 171391, 171713 }, book = 176089 }, -- leatherworking
  [197] = { iconid = 111556, spells = { 168835, 169669 }, book = 176058 }, -- tailoring
}
PT.professionCooldownSpellIDs = {}
PT.formatStrings = {}
PT.formatStrings.iconTexture = "|T%s:16:16:0:0:64:64:4:60:4:60|t"
function PT:LevelCheck(info, newMinLevel)
  self.db.profile.minLevel = newMinLevel
end

function PT:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
  self.db = LibStub("AceDB-3.0"):New("ProfessionTrackDB", defaults, true)
  self:MigrateSavedData()

  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, self.GenerateOptions)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName)

  PTLDB = ldb:NewDataObject(addonName, {
    type = "data source",
    text = addonName,
    icon = "Interface\\Icons\\inv_eng_gearspringparts",
    OnClick = function(clickedframe, button)
      if button == "RightButton" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
      else
        self:UpdatePlayer()
        self:UpdateItemCounts()
        self:UpdateProfessionCooldowns()
        self:UpdateLDB()
      end
    end,
    OnEnter = function(tt)
      local tooltip = LibQTip:Acquire("PTTooltip", 1, "LEFT")
      self.tooltip = tooltip
      self:PrepareTooltip(tooltip)
      tooltip:SmartAnchorTo(tt)
      tooltip:Show()
    end,
    OnLeave = function(tt)
      LibQTip:Release(PT.tooltip)
      self.tooltip = nil
    end,
  })

  LDBIcon:Register(addonName, PTLDB, self.db.profile.minimapIcon)


  currentRealmChar = strformat("%s.%s", GetRealmName(), UnitName("player"))
  if not PT.db.global.cooldowns[currentRealmChar] then
    PT.db.global.cooldowns[currentRealmChar] = {}
  end

  -- Generate a list of all spellids we want to listen to
  local p, s
  for _, p in pairs(self.professionCooldowns) do
    for _, s in pairs( p.spells ) do
      self.professionCooldownSpellIDs[s] = true
    end
    self.professionCooldownSpellIDs[p.book] = true
  end

  C_Timer.After( 0.1, function() self:GenerateProfessionTextures() end )
end

function PT:GenerateProfessionTextures()
  local done, itemTexture = true
  for _, p in pairs(self.professionCooldowns) do
    itemTexture = GetItemIcon( p.iconid )
    if not itemTexture then
      done = false
    else
      p.texture = strformat(self.formatStrings.iconTexture, itemTexture)
    end
  end

  -- Repeat if something was not loaded properly
  if not done then
    C_Timer.After( 0.1, function() self:GenerateProfessionTextures() end )
  end
end

-- Options Handling

function PT:GenerateOptions()
  if not PT.options then

    local function newCounter()
      local i = 0
      return function( inc )
          i = i + 1
          return i
        end
    end
    local count = newCounter()

    PT.options = {
      name = addonName,
      type = 'group',
      handler = PT,
      get = function(info) return PT.db.profile[info[#info]] end,
      set = function(info, value) PT.db.profile[info[#info]] = value end,
      args = {
        genconfig = {
          name = L["General Settings"],
          type = 'group',
          order = 1,
          args = {
            usedcd = {
              name = L["Used CD color"],
              desc = L["Used CD color description"],
              type = 'color',
              order = count(),
              get  = getColor,
              set  = setColor,
              hasAlpha = true,
            },
            freecd = {
              name = L["Free CD color"],
              desc = L["Free CD color description"],
              type = 'color',
              order = count(),
              get  = getColor,
              set  = setColor,
              hasAlpha = true,
            },
            tooltipBGColor = {
              name = L["Background color"],
              desc = L["Background color"],
              type = 'color',
              order = count(),
              get  = function(info)
                return PT.db.profile[info[#info]].r, PT.db.profile[info[#info]].g, PT.db.profile[info[#info]].b, PT.db.profile[info[#info]].a
              end,
              set  = function(info, r, g, b, a)
                PT.db.profile[info[#info]] = { r = r, g = g, b = b, a = a }
              end,
              hasAlpha = true,
            },
            useClassColors = {
              type = "toggle",
              order = count(),
              name = L["Color names by class"],
            },
            useCustomClassColors = {
              type = "toggle",
              order = count(),
              name = L["Use !ClassColors"],
              desc = L["Use !ClassColors addon for class colors used to color the names in the tooltip"], 
              set = function(info, value)
                PT.db.profile[info[#info]] = value
                PT:SetupClasscoloredFonts()
              end,
              disabled = function()
                return not PT.db.profile.useClassColors or not CUSTOM_CLASS_COLORS
              end,
            },
            showRealmHeaderLines  = {
              type = "toggle",
              order = count(),
              name = L["Show Realm Headers"],
              desc = L["SHOW_REALMLINES_DESC"],
            },
            showShipments = {
              type = "toggle",
              order = count(),
              name = L["Add 'Broker: Garrison' shipment data"],
              desc = L["If the addon 'Broker: Garrison' is installed then use its data to incorporate finished work orders into the tooltip"],
              disabled = not BrokerGarrison,
            },
            showRealmSumLines  = {
              type = "toggle",
              order = count(),
              name = L["Show row with sum"],
              desc = L["SHOW_SUMROW_DESC"],
            },
            showCooldowns = {
              type = "toggle",
              order = count(),
              name = L["Show cooldowns"],
              desc = L["Show columns for profession cooldowns"],
            },
            showBooks = {
              type = "toggle",
              order = count(),
              name = L["Show a 'b' for 'Secref of ...'"],
              desc = L["Adds a 'b' to the cooldown column if the book-spell 'Secret of draenor ...' is on cooldown. (Will only show if the main spell is on cooldown too!)"],
            },

            showLDBCooldownsAvailable = {
              type = "toggle",
              order = count(),
              name = L["LDB: Show Cooldowns"],
              desc = L["Add the number of available cooldoewns to the LDB text"],
              set = function(info, value)
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            showLDBCooldownsText = {
              type = "input",
              order = count(),
              name = L["LDB: Cooldowns text"],
              desc = L["Formatstring for the text, use %d for the number of cooldowns"],
              set = function(info, value)
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            showLDBItemsAvailable = {
              type = "toggle",
              order = count(),
              name = L["LDB: Show Items"],
              desc = L["Add the number of profession items on the current char to the LDB text"],
              set = function(info, value)
                      PT.db.profile[info[#info]] = value
                      PT:UpdateLDB()
                    end,
            },
            minimapIcon = {
              type = "toggle",
              name = L["Show minimap button"],
              desc = L["Show the PT minimap button"],
              order = count(),
              get = function(info) return not PT.db.profile.minimapIcon.hide end,
              set = function(info, value)
                PT.db.profile.minimapIcon.hide = not value
                if value then
                  LDBIcon:Show(addonName)
                else
                  LDBIcon:Hide(addonName)
                end
              end,
            },
          },
        },
        databasemanagement = {
          name = L["Data"],
          type = 'group',
          order = 2,
          args = {
            minLevel = {
              type = "range",
              order = count(),
              min = 1;
              max = MAX_PLAYER_LEVEL_TABLE[GetExpansionLevel()];
              step = 1,
              name = L["Minimum Level"],
              desc = L["Characters below this level will not be shown"],
              set = "LevelCheck",
            },
          },
        },
        tracking = {
          name = L["Tracked Items"],
          type = 'group',
          order = 3,
          args = {},
        },
      },
    }

    -- Tracking options helper functions
    local function RemoveTrackedItemFromOptions( itemID )
      DebugPrint("Removing", itemID)
      -- Remove Item from tracking and the correpsonding option entries
      local globaldb = PT.db.global
      globaldb.trackedItems[itemID] = false

      PT.options.args.tracking.args[tostring(itemID)] = nil
      PT.options.args.tracking.args[tostring(itemID)..".1"] = nil
      PT.options.args.tracking.args[tostring(itemID)..".0"] = nil
      PT.options.args.tracking.args[tostring(itemID).."newline"] = nil

      -- Find the itemID in the sort list, make the one before point to the one
      -- after then delete itemID from the list
      local current, found = 1
      local alreadyseen = {}
      while globaldb.trackedSort[current] and not found and not alreadyseen[current] do
        alreadyseen[current] = true
        if globaldb.trackedSort[current] == itemID then
          found = current
        end
        current = globaldb.trackedSort[current]
      end
        DebugPrint("Removing #2", current, found)
      if not found then
        DebugPrint("Removing", itemID,  "not found")
      else 
        -- ... found -> itemID -> nextItem ...
        -- => ... found -> nextItem ...
        DebugPrint("Removing", itemID, found, "->", globaldb.trackedSort[found], "->", globaldb.trackedSort[itemID])
        globaldb.trackedSort[found] = globaldb.trackedSort[itemID]
        globaldb.trackedSort[itemID] = nil
      end

      -- Updates LDB display
      PT:UpdateItemCounts()
      PT:UpdateLDB()
    end


    local function MoveItemDownTheList( listPlace )
      -- x -> y: x points to y
      -- Current:   listPlace -> item1 -> item2 -> restOfItems
      -- Change to: listPlace -> item2 -> item1 -> restOfItems
      local trackSortList = PT.db.global.trackedSort
      local swap1       = trackSortList[listPlace]
      local swap2       = trackSortList[swap1]
      local restOfItems = trackSortList[swap2]
      local optionstracking = PT.options.args.tracking
      -- Only move down if we are not the last item on the list
      if listPlace and listPlace > 0 and swap1 > 0 and swap2 > 0 then
        DebugPrint( listPlace .." -> ".. swap1 .." -> ".. swap2 .. " -> "..restOfItems )

        local tmporder = {
          optionstracking.args[ tostring( swap1 ) ].order,
          optionstracking.args[ tostring( swap1 ) ..".1"].order,
          optionstracking.args[ tostring( swap1 ) ..".0"].order,
        }

        optionstracking.args[ tostring( swap1 ) ].order       = optionstracking.args[ tostring( swap2 ) ].order
        optionstracking.args[ tostring( swap1 ) ..".1"].order = optionstracking.args[ tostring( swap2 ) ..".1"].order
        optionstracking.args[ tostring( swap1 ) ..".0"].order = optionstracking.args[ tostring( swap2 ) ..".0"].order
        optionstracking.args[ tostring( swap2 ) ].order       = tmporder[1]
        optionstracking.args[ tostring( swap2 ) ..".1"].order = tmporder[2]
        optionstracking.args[ tostring( swap2 ) ..".0"].order = tmporder[3]

        -- Current:
        -- trackSortList[listPlace] = swap1
        -- trackSortList[swap1]     = swap2
        -- trackSortList[swap2]     = restOfItems
        -- Change to:
        trackSortList[listPlace] = swap2
        trackSortList[swap2]     = swap1
        trackSortList[swap1]     = restOfItems
        DebugPrint( listPlace .." -> ".. trackSortList[listPlace] .." -> ".. trackSortList[trackSortList[listPlace]] .. " -> "..trackSortList[trackSortList[trackSortList[listPlace]]] )
        PT:UpdateLDB()
      else
        DebugPrint( ( listPlace ).." is last on the list" )

      end
    end
    local function MoveItemUpwards( itemIDstring )
      -- This is basically moving the item above downwards
      DebugPrint( "MoveItemUpwards("..itemIDstring..")" )
      local itemID, _ = strmatch(itemIDstring, "(.+)%.(.+)")
      itemID = tonumber(itemID)
      local current, globaldb, found = 1, PT.db.global

      -- Make sure we are not already the top item
      if globaldb.trackedSort[current] == itemID then
        DebugPrint( "Tried to move the first item upwards!" )
      else
        --   a -> ... itemX -> itemY -> itemToMoveUpwards -> rest-of-items ...
        -- change this to
        --   a -> ... itemX -> itemToMoveUpwards -> itemY -> rest-of-items ...
        -- We just need to find itemX and do MoveItemDownTheList( itemX )
        -- which will make the two following items swap places


        while globaldb.trackedSort[current] and not found do
          -- When moving upwards we need to look ahead twice
          if not globaldb.trackedSort[current] and
             not globaldb.trackedSort[ globaldb.trackedSort[current] ] then
             -- Have not found it by now and we are at the end of the list
             -- So the item we want to move seems to be not even on it!
            DebugPrint( "MoveItemUpwards: "..itemID.." not found in trackedSort")
          end

          if globaldb.trackedSort[ globaldb.trackedSort[current] ] == itemID then
            -- Found the items two steps ahead of us, so if we swap the next
            -- two we actually move the one we want up a notch
            found = true
          else
            current = globaldb.trackedSort[current]
          end
        end
        if not found then
          DebugPrint("This should never happen, tried to move "..itemID.." upwards but could not find it!")
        else
          MoveItemDownTheList( current )
        end
      end
    end

    local function MoveItemDownwards( itemIDstring )
      DebugPrint( "MoveItemDownwards("..itemIDstring..")" )
      local itemID, _ = strmatch(itemIDstring, "(.+)%.(.+)")
      itemID = tonumber(itemID)
      local current, globaldb, found = 1, PT.db.global
      -- Need to find which the item pointing towards the one we want to move
      -- down
      while globaldb.trackedSort[current] and not found do
        if globaldb.trackedSort[current] == itemID then
          found = true
        else
          current = globaldb.trackedSort[current]
        end
      end
      if not found then
        DebugPrint("This should never happen, tried to move "..itemID.." downwards but could not find it!")
      else
        MoveItemDownTheList( current )
      end
    end
    local rarityColors = {}
    for i=0,7 do
      rarityColors[i] = "|c"..select(4,GetItemQualityColor(i)).."%s|r"
    end
    local function AddTrackedItemToOptions( item )

      local itemID
      local iname, link, iRarity = GetItemInfo( item )
      if not link then
        return
      else
        _, _, _, _, itemID, _, _, _, _, _, _, _, _, _  = string.find(link, "|?c?f?f?(%x*)|?H?([^:]*):?(%d+):?(%d*):?(%d*):?(%d*):?(%d*):?(%d*):?(%-?%d*):?(%-?%d*):?(%d*):?(%d*):?(%-?%d*)|?h?%[?([^%[%]]*)%]?|?h?|?r?")
      end
      DebugPrint("Adding ".. iname.. " (".. itemID.. ") to tracking")
      itemID = tonumber(itemID)
      local globaldb = PT.db.global
      -- Some sanity checks on the added item
      if itemID and itemID > 0 and iname and not PT.options.args.tracking.args[tostring(itemID)] then
        globaldb.trackedItems[itemID] = true
        PT.options.args.tracking.args[tostring(itemID)]  = {
          type = "execute",
          name = strformat(rarityColors[iRarity], iname),
          image = GetItemIcon( itemID ),
          func = function(info)
            RemoveTrackedItemFromOptions( tonumber(info[#info]) )
          end,
          order = count()
        }
        PT.options.args.tracking.args[tostring(itemID)..".1"]  = {
          type = "execute",
          name = "^",
          width = "half",
          hidden = (not debug),
          func = function(info)
            MoveItemUpwards(info[#info])
          end,
          order = count()
        }
        PT.options.args.tracking.args[tostring(itemID)..".0"]  = {
          type = "execute",
          name = "v",
          width = "half",
          hidden = (not debug),
          func = function(info)
            MoveItemDownwards(info[#info])
          end,
          order = count()
        }
        PT.options.args.tracking.args[tostring(itemID).."newline"] = {
          type = "description",
          name = "",
          width = "full",
          order = count(),
        }
        -- Just add the item to the end of the list for now
        local current = 1
        while globaldb.trackedSort[current] and globaldb.trackedSort[current] > 0 do
          current = globaldb.trackedSort[current]
        end
        DebugPrint("Last item: ".. current.. " adding ".. itemID.. " behind")
        if not globaldb.trackedSort[itemID] then
          -- If it's not in there we just added a new item, add it to the end
          -- Else it was one that was already tracked
          globaldb.trackedSort[current] = itemID
          globaldb.trackedSort[itemID]  = 0
        end
        PT:UpdateItemCounts()
        PT:UpdateLDB()
      end
    end

    PT.options.args.tracking.args.adddefault = {
      type = "execute",
      name = L["Add Default Items"],
      func = function(info)
        local trackid
        for trackid, _ in pairs(defaults.global.trackedItems) do
          AddTrackedItemToOptions(trackid)
        end
      end,
      order = count()
    }

    PT.options.args.tracking.args.addOreHerbsLeather = {
      type = "execute",
      name = L["Add Ore/Herbs/Leather"],
      width = "full",
      func = function(info)
        -- Mining
        AddTrackedItemToOptions(123918) -- Leystone
        AddTrackedItemToOptions(123919) -- Felslate
        AddTrackedItemToOptions(124444) -- Infernal Brimstone
        -- Herblism
        AddTrackedItemToOptions(124101) -- Aethril
        AddTrackedItemToOptions(124102) -- Dreamleaf
        AddTrackedItemToOptions(124103) -- Foxflower
        AddTrackedItemToOptions(124104) -- Fjarnskaggl
        AddTrackedItemToOptions(124105) -- Starlight Rose
        AddTrackedItemToOptions(124106) -- Felwort
        -- Skinning
        AddTrackedItemToOptions(124113) -- Stonehide Leather
        AddTrackedItemToOptions(136533) -- Dreadhide Leather
        AddTrackedItemToOptions(124115) -- Stormscale
        AddTrackedItemToOptions(136534) -- Gravenscale
        AddTrackedItemToOptions(124116) -- Felhide
      end,
      order = count()
    }

    PT.options.args.tracking.args.addbloodofsargeras = {
      type = "execute",
      name = L["Add Blood of Sargeras"],
      func = function(info)
        AddTrackedItemToOptions(124124)
      end,
      order = count()
    }

    PT.options.args.tracking.args.itemaddinput = {
      type = "input",
      name = L["Add Item"],
      get = function( info ) return "" end,
      set = function( info, value )
        AddTrackedItemToOptions(value)
      end,
      order = count()
    }

    PT.options.args.tracking.args.trackingheader = {
      type = "header",
      name = L["Items"],
      order = count()
    }

    DebugPrint("Adding tracked items to options")
    local current, globaldb = 1, PT.db.global
    local alreadyseen = {}
    while globaldb.trackedSort[current] > 0 and not alreadyseen[current] do
      alreadyseen[current] = true
      current = globaldb.trackedSort[current]
      if globaldb.trackedItems[current] then
        AddTrackedItemToOptions( current )
      end
    end
    local sortedRoster, lastRealm = {}
    for realmChar, _ in pairs(globaldb.roster) do
      tinsert(sortedRoster, realmChar)
    end
    table.sort(sortedRoster)

    function colorName(name, classColor)
    	return strformat("|cff%02x%02x%02x%s|r",classColor.r*255,classColor.g*255,classColor.b*255, name)
    end
    for _, realmChar in pairs(sortedRoster) do
      local currRealm, currChar = strmatch(realmChar, "(.+)%.(.+)")
      local charData = globaldb.roster[realmChar]
      if lastRealm ~= currRealm then
        -- Realm Header --
        lastRealm = currRealm
        PT.options.args.databasemanagement.args["header"..currRealm] = {
          type="header",
          name=currRealm,
          order = count(),
        }
      end
      PT.options.args.databasemanagement.args[realmChar] = {
        type = "description",
        width = "normal",
        name = colorName(currChar, RAID_CLASS_COLORS[charData.class]),
        order = count(),
      }
--[[
      PT.options.args.databasemanagement.args[realmChar.."ignore"] = {
        type = "toggle",
        width = "half",
        get = function(info) return PT.db.global.roster[realmChar].ignored end,
        set = function(info, value) PT.db.global.roster[realmChar].ignored = value end,
        name = L["Ignore"],
        order = count(),
      } --]]--
      local val, cdid = { [-1] = "All", [0] = "None" }
      for cdid, _ in pairs(PT.db.global.cooldowns[realmChar]) do
        val[cdid] = PT.professionCooldowns[cdid].texture
      end

      PT.options.args.databasemanagement.args[realmChar.."ignore"] = {
        type = "select",
        width = "half",
        values = val,
        get = function(info) return PT.db.global.roster[realmChar].ignored end,
        set = function(info, value) 
                PT.db.global.roster[realmChar].ignored = value
                PT:UpdateLDB()
              end,
        name = L["Ignore"],
        order = count(),
      }
      PT.options.args.databasemanagement.args[realmChar.."delete"] = {
        type = "execute",
        width = "half",
        name = L["Delete"],
        func = function(info)
          PT:DeleteCharacterData(realmChar)
        end,
        order = count(),
      }
      PT.options.args.databasemanagement.args[realmChar.."newline"] = {
        type = "description",
        name = "",
        width = "full",
        order = count(),
      }
    end


    PT.options.args.profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(PT.db)
  end
  return PT.options
end


---- Tooltip handling/helpers
function PT:PrepRosterSlice()
  -- returns a roster filtered by level check into which resets can be added
  local minLevel = self.db.profile.minLevel
  local globaldb = self.db.global
  local roster = {}

  for realmChar, data in pairs(globaldb.roster) do
    if minLevel <= data.level and ( not data.ignored or ( data.ignored and data.ignored > -1 )) then
      local realm, char = strmatch(realmChar, "(.+)%.(.+)")
      roster[realmChar] = {
        name = char,
        level = data.level,
        color = RAID_CLASS_COLORS_FONTS[data.class],
        items = globaldb.items[realmChar],
        cooldowns = globaldb.cooldowns[realmChar],
        shipments = self:GetBrokerGarrisonShipementData(realm, char),
        ignored = data.ignored
      }
    end
  end

  -- structure is { realmChar = { (char)name, level, color, items = { name = data }, cooldowns = { name = data } } }
  return roster
end

function PT:SortRosterSlice(roster)
  -- roster param structure is { realmChar = { (char)name, level, color, items = { name = data }, cooldowns = { name = data } } }
  -- basically, turn the key into sorted tables which point to the value
  local sortedRoster = {}
  local temp = {}

  for realmChar,data in pairs(roster) do
    tinsert(temp, realmChar)
  end
  tsort(temp)

  local curRealm
  for _,realmChar in ipairs(temp) do
    local realm, char = strmatch(realmChar, "(.+)%.(.+)")

    if curRealm ~= realm then
      curRealm = realm
      tinsert(sortedRoster, { name = realm, chars = {} })
    end

    tinsert( sortedRoster[#sortedRoster].chars, roster[realmChar] )
  end

  -- structure is { [iRealm] = { name, chars = { [iChar] = { name, level, color, items = { name = data }, cooldowns = { name = data } } } } }
  return sortedRoster
end


PT.buildingID2ItemID = {
  -- The Forge
  [60]  = { item =108257, mul = 1 },
  [117] = { item =108257, mul = 2 },
  [118] = { item =108257, mul = 2 },
  -- Alchemy Lab
  [76]  = { item =108996, mul = 1 },
  [119] = { item =108996, mul = 2 },
  [120] = { item =108996, mul = 2 },
  -- The Tannery
  [90]  = { item =110611, mul = 1 },
  [121] = { item =110611, mul = 2 },
  [122] = { item =110611, mul = 2 },
  -- Engineering Works
  [91]  = { item =111366, mul = 1 },
  [123] = { item =111366, mul = 2 },
  [124] = { item =111366, mul = 2 },
  -- Tailoring Emporium
  [94]  = { item =111556, mul = 1 },
  [127] = { item =111556, mul = 2 },
  [128] = { item =111556, mul = 2 },
  -- Sribe's Quarters
  [95]  = { item =112377, mul = 1 },
  [129] = { item =112377, mul = 2 },
  [130] = { item =112377, mul = 2 },
  -- Gem Boutique
  [96]  = { item =115524, mul = 1 },
  [131] = { item =115524, mul = 2 },
  [132] = { item =115524, mul = 2 },
  -- Enchanter's Study
  [93] =  { item = 115504, mul = 1 },
  [125] = { item = 115504, mul = 2 },
  [126] = { item = 115504, mul = 2 },
}

function PT:GetBrokerGarrisonShipementData(server, character)
  local shipments = {}
  if not self.db.profile.showShipments or not BrokerGarrison then
    return shipments
  end
  local garrisonData =  BrokerGarrison.get(server, character)
  if not garrisonData then
    return shipments
  end
  for _, buildingData in pairs(garrisonData.buildings) do
    if self.buildingID2ItemID[buildingData.id] then
      local rdy, est = buildingData.shipment.shipmentsReady or 0, buildingData.shipment.shipmentsReadyEstimate or 0
      local count = est > rdy and est or rdy
      -- Only use multiplier if a follower is in the building, can't know for
      -- how many orders it was in there though
      local mul = buildingData.follower and next(buildingData.follower) and self.buildingID2ItemID[buildingData.id].mul or 1
      DebugPrint(server, character, buildingData.shipment.name, count, mul)
      shipments[self.buildingID2ItemID[buildingData.id].item] = count * mul
    end
  end
  return shipments
end

function PT:SetupClasscoloredFonts()
  local class, color, CHOSEN_CLASS_COLORS

  if self.db.profile.useCustomClassColors and CUSTOM_CLASS_COLORS then
    CHOSEN_CLASS_COLORS = CUSTOM_CLASS_COLORS
  else
    CHOSEN_CLASS_COLORS = RAID_CLASS_COLORS
  end
  for class,color in pairs(CHOSEN_CLASS_COLORS) do
    if not RAID_CLASS_COLORS_FONTS[class] then
      RAID_CLASS_COLORS_FONTS[class] = CreateFont("ClassFont"..class)
      RAID_CLASS_COLORS_FONTS[class]:CopyFontObject(GameTooltipText)
    end
    RAID_CLASS_COLORS_FONTS[class]:SetTextColor(color.r, color.g, color.b)
  end
end



function PT:PrepareTooltip(tooltip)

  tooltip:SetBackdropColor(self.db.profile.tooltipBGColor.r, self.db.profile.tooltipBGColor.g, self.db.profile.tooltipBGColor.b, self.db.profile.tooltipBGColor.a)
  --[[  Each char can have up to 2 cooldowns, but
        any kind of item in the inventroy thanks
        to garrison buildings

        cd1 cd2 i1  i2  i3  i4  i5  i6  i7  i8
  Char1 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  Char2 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  Char3 [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ] [ ]
  -----]]--
  local Columns = {}
  local function MyAddColumn(nameType, displayName)
    if nameType then
      if not displayName then
        DebugPrint("MyAddColumn: missing displayName for", nameType)
        displayName = nameType
      end
      if not formatFunc then
        --formatFunc = CellFormat.Default
      end

      if not Columns[nameType] then
        local colindex = tooltip:AddColumn("CENTER")
        tooltip:SetCell(2, colindex, displayName)

        Columns[nameType] = { col = colindex }
      else
        DebugPrint("a column already exists for", nameType)
      end
    end
  end
  local strResetBreaker = "(([^%.]+)%.([^%.]+))%.([^%.]+)"   -- returns realmChar, realm, char, name

  local globaldb = self.db.global
  local profiledb = self.db.profile

  local roster = self:PrepRosterSlice()  -- structure is { realmChar = { (char)name, level, color, items = { [id] = count }, cooldowns = {} } }
  local found = {}

  local id, show, itemTexture, formatedTexture

  tooltip:AddHeader()
  tooltip:AddHeader()

  if profiledb.showCooldowns then
    MyAddColumn( "cd1", "CD1")
    MyAddColumn( "cd2", "CD2")
  end


  local current = 1
  while globaldb.trackedSort[current] and globaldb.trackedSort[current] > 0 do
    current = globaldb.trackedSort[current]
    local itemTexture = GetItemIcon( current )
    local formatedTexture = strformat(self.formatStrings.iconTexture, itemTexture)
    MyAddColumn( current, formatedTexture )
  end





  for dbkey,data in pairs(globaldb.items) do
    local realmChar, realm, char, name = strmatch(dbkey, strResetBreaker)
    if roster[realmChar] then
      found[nameType] = true
    end
  end

  for dbkey,data in pairs(globaldb.cooldowns) do
    local realmChar, realm, char, name = strmatch(dbkey, strResetBreaker)
    if roster[realmChar] then
      found[nameType] = true
    end
  end

  local sortedRoster = self:SortRosterSlice( roster )

  local realmItemsSum = {}
  for _, realm in ipairs(sortedRoster) do
    if profiledb.showRealmHeaderLines then
      tooltip:SetCell(tooltip:AddLine(), 1, realm.name, nil, "CENTER", tooltip:GetColumnCount())
    end
    realmItemsSum[realm.name] = {}


    for _, currChar in ipairs(realm.chars) do
      -- local charRealm = char.." - "..realm
      local lineName = currChar.name -- profile.showCharacterRealm and charRealm or char
      local lineFont = profiledb.useClassColors and currChar.color or nil --  RAID_CLASS_COLORS_FONTS[ self.db.global.charClass[charRealm] ] or nil
      local row = tooltip:AddLine("")

      tooltip:SetCell(row, 1, lineName, lineFont)

      for itemid, show in pairs( globaldb.trackedItems ) do
        if not realmItemsSum[realm.name][itemid] then
          realmItemsSum[realm.name][itemid] = 0
        end
        if show then
          tooltip:SetCell(row, Columns[itemid].col, ( currChar.items[itemid] or 0 ).. ( currChar.shipments[itemid] and currChar.shipments[itemid]>0 and ( "+"..currChar.shipments[itemid] ) or "" ) )
          realmItemsSum[realm.name][itemid] = realmItemsSum[realm.name][itemid] + ( currChar.items[itemid] or 0 ) + ( currChar.shipments[itemid] or 0 )
        end

      end
      local now = time()
      local cdused, cellContent
      if profiledb.showCooldowns then
        local cd = 1
        for skillLine, data in pairs(currChar.cooldowns) do
          if skillLine ~= currChar.ignored then
            if cd > 2 then
              DebugPrint("Should not happen: cd="..cd)
            end
            cellContent = self.professionCooldowns[skillLine].texture or "X"
            if data.reset and data.reset > now then
              if profiledb.showBooks and data.bookReset and data.bookReset > now then
                cellContent = cellContent.."b"
              end
              tooltip:SetCell(row, Columns["cd"..cd].col, cellContent)
              tooltip:SetCellColor(row, Columns["cd"..cd].col, profiledb.usedcd.r, profiledb.usedcd.g, profiledb.usedcd.b, profiledb.usedcd.a)
            else
              tooltip:SetCell(row, Columns["cd"..cd].col, cellContent)
              tooltip:SetCellColor(row, Columns["cd"..cd].col, profiledb.freecd.r, profiledb.freecd.g, profiledb.freecd.b, profiledb.freecd.a)
            end
            cd = cd+1
          end
        end
      end
    end
    if profiledb.showRealmSumLines then
      local row = tooltip:AddLine("")
      tooltip:SetCell(row, 1, L["SUM"], nil, "RIGHT" )
      for itemid, show in pairs( globaldb.trackedItems ) do
        if show then
          tooltip:SetCell(row, Columns[itemid].col, realmItemsSum[realm.name][itemid] or 0 )
        end
      end
    end -- profiledb.showRealmSumLines
  end -- realm in ipairs(sortedRoster)
end

function PT:UpdateLDB()
  local LDBcd, LDBitem
  local roster = self:PrepRosterSlice()
  if self.db.profile.showLDBCooldownsAvailable then
    local availableCooldowns = 0
    for realmChar, chardata in pairs(roster) do
      if chardata.level > 89 then
        for prof, profdata in pairs(chardata.cooldowns) do
          -- (whole character not ignored and profession not ignored AND off cooldown?
          if ( chardata.ignored  and chardata.ignored > -1 and chardata.ignored ~= prof ) and
             ( not profdata.reset or ( profdata.reset and profdata.reset < time() )) then
            DebugPrint(realmChar, prof, time(), profdata.reset or 0, chardata.ignored)
            availableCooldowns = availableCooldowns + 1
          end
        end
      end
    end
    LDBcd = strformat(self.db.profile.showLDBCooldownsText, availableCooldowns)
  end
  if self.db.profile.showLDBItemsAvailable then
    local tmptable = {}
    local globaldb, itemTexture = self.db.global
    local current = 1
    local alreadyseen = {}
    while globaldb.trackedSort[current] and globaldb.trackedSort[current] > 0 and not alreadyseen[current] do
      alreadyseen[current] = true
      current = globaldb.trackedSort[current]
      if globaldb.items[currentRealmChar][current] > 0  and self.db.global.trackedItems[current] then
        itemTexture = strformat(self.formatStrings.iconTexture,GetItemIcon( current ) )
        tinsert(tmptable, globaldb.items[currentRealmChar][current]..itemTexture)
      end
    end

    if #tmptable > 0 then
      LDBitem = table.concat(tmptable, ",")
    end
  end
  PTLDB.text = ""..(LDBcd and LDBcd or "")..( ( LDBcd and LDBitem ) and "," or "")..(LDBitem and LDBitem or "")
end



--

function PT:OnEnable()
  self:RegisterEvent("BAG_UPDATE",               "UpdateItemCounts")
  self:RegisterEvent("PLAYER_LEVEL_UP",          "UpdatePlayer")
  self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "CheckForProfessionSpell")

  self:SetupClasscoloredFonts()
  self:UpdatePlayer()
  self:UpdateItemCounts()
  self:UpdateProfessionCooldowns()
end

function PT:OnDisable()
  self:UnregisterEvent("BAG_UPDATE")
  self:UnregisterEvent("PLAYER_LEVEL_UP")
  self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
end


---- Handling Tracking

function PT:UpdatePlayer()
  local level = UnitLevel("player")
  local class = select(2,UnitClass("player"))
  if not self.db.global.roster[currentRealmChar] then 
    self.db.global.roster[currentRealmChar] = {}
  end
  self.db.global.roster[currentRealmChar].class = class
  self.db.global.roster[currentRealmChar].level = level
end

function PT:CheckForProfessionSpell( ... )
  local spellID = select(6, ...)
  if self.professionCooldownSpellIDs[spellID] then
    C_Timer.After( 0.5, function() self:UpdateProfessionCooldowns() end )
  end
end


function PT:UpdateProfessionCooldowns()
  if self:GetNextDailyReset() then
    wipe( PT.db.global.cooldowns[currentRealmChar] )
    local p1, p2, skillLine
    p1, p2 = GetProfessions()
    if p1 then
      skillLine = select(7, GetProfessionInfo( p1 ))
      if self.professionCooldowns[skillLine] then
        self:UpdateProfessionCooldown( skillLine )
      end
    end
    if p2 then
      skillLine = select(7, GetProfessionInfo( p2 ))
      if self.professionCooldowns[skillLine] then
        self:UpdateProfessionCooldown( skillLine )
      end
    end
    self:UpdateLDB()
  else
    C_Timer.After( 0.5, function() self:UpdateProfessionCooldowns() end )
  end
end

function PT:UpdateProfessionCooldown( skillLine )
  if not PT.db.global.cooldowns[currentRealmChar][skillLine] then
    PT.db.global.cooldowns[currentRealmChar][skillLine] = {}
  end

  local nextReset = self:GetNextDailyReset()
  for _, checkspell in pairs( self.professionCooldowns[skillLine].spells ) do
    if GetSpellCooldown(checkspell) > 0 then
      PT.db.global.cooldowns[currentRealmChar][skillLine].reset = nextReset
    end
  end
  if GetSpellCooldown( self.professionCooldowns[skillLine].book ) > 0 then
    PT.db.global.cooldowns[currentRealmChar][skillLine].bookReset = nextReset
  end
end

function PT:UpdateItemCounts()
  local globaldb, key, id  = self.db.global
  if not globaldb.items[currentRealmChar] then
    globaldb.items[currentRealmChar] = {}
  end
  for id, _ in pairs(globaldb.trackedItems) do
    globaldb.items[currentRealmChar][id] = GetItemCount( id, true )
  end
  self:UpdateLDB()
end

---- Database Maintenance

function PT:MigrateSavedData()
  local  globaldb = self.db.global
  local profiledb = self.db.profile

  -- all db sections are current we can exit immediately
  if profiledb.version == DB_VERSION and
      globaldb.version == DB_VERSION then
    return
  else
    PT:Print(L["DB_VERSION_UPGRADE_PURGE"])
  end

  -- Migrating from before addon began keeping track of database changes
  --[[
  if globaldb.version < X then
    -- Upgrade from before version X
  end
  if profiledb.version < X then
    -- Upgrade from before version X
  end
  --]]--

  profiledb.version = DB_VERSION
  globaldb.version = DB_VERSION
end

function PT:DeleteCharacterData(realmChar)
  self.options.args.databasemanagement.args[realmChar]            = nil
  self.options.args.databasemanagement.args[realmChar.."ignore"]  = nil
  self.options.args.databasemanagement.args[realmChar.."delete"]  = nil
  self.options.args.databasemanagement.args[realmChar.."newline"] = nil
  self.db.global.items[realmChar]     = nil
  self.db.global.roster[realmChar]    = nil
  self.db.global.cooldowns[realmChar] = nil
  self.db.global.items[realmChar]     = nil
end

----


function PT:GetNextDailyReset()
  -- when first logging in, we may get a bogus result from GetQuestResetTime
  -- if thats the case, then just return nil which triggers the update
  -- functions to requeue 0.5s later
  local qrtime = GetQuestResetTime()
  if qrtime <= 0 then
    return nil
  end
  return time() + qrtime
end

